#include "ddf.h"
#include "io.h"
#include "files.h"

using std::string;

int main(int argc, char* argv[])  //reads command line inputfile, output.
{
  /* Initialize input, output files.*/
  initializefiles(argc,argv);
  std::ifstream inputfile(argv[1]);
  std::ofstream outputfile(argv[2]); //files=streams

  XVIII_Decomp good_circ; /* Data for 18 gate circuit; see q-ph/0303038.*/
  FourByFour v=Read_4x4 (inputfile ); /* Loads v from input file.*/
  v=GramSchmidt(v); /*v should be unitary.Apply GS to be sure. See diag_4x4.h*/
  if( complex_size(det(v)) < 0.1) /* Ooops! */
    {
      std::cerr << "ERROR! G-Schmidt of input matrix is not unitary." << endl;
      return -1; /* Leave int main(...) in error state.*/
    }
  else
    {
      /* Step over all Canonical decomps looking for small 3 CNOT circ.*/
      good_circ = can_dec_opt_circ (v); 
      
      string output;
      /* Write output to string.*/
      output += LaTeX_header();
      output += Print_U( v, argv[1]);
      output += Print_Circ ( good_circ);
      output += Print_Conventions();
  
      /*  Write string to output file.*/
      outputfile << output;
    };
  return 0;
}
