/*  Functions for stepping over discrete degrees of freedom in two-qubit   */
/*  canonical decompositions.                                              */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Standards and  Technology,                        */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */




#ifndef _files_H_
#define _files_H_

#include<iostream>
#include<fstream>
#include <cstdlib>
#include "twoQbit.h"

/* Prints error messages.*/
void printerror (const char* p, const char* p2="")
{
  std::cerr << p << ' ' << p2<<std::endl;
  std::exit(1);
};

/* Initialize files as stream.  printerror if initialization fails.        */
void initializefiles(int argc, char* argv[])
{
  if(argc != 3) 
    printerror("Program expects \"./3_cnot_synth foo1.txt foo2.tex\".\n foo1.txt input unitary. (See randUNI.txt for format.)\n foo2.tex is LaTeX source output by program.\n");
 
  std::ifstream from(argv[1]);  //opens input file stream
  if(!from) printerror ("can not open input file ",argv[1]);

  std::ofstream to(argv[2]);  //opens output file stream
  if(!to) printerror("can not open output file ",argv[2]);
};

/* Loads 4x4 matrix of complexes from stream (e.g. input file.)           */
FourByFour Read_4x4 (istream& in)
{
  double a,b; char ch; complex z;
  FourByFour v;

  for(int j=1;j<=4;++j)
      for(int k=1;k<=4;++k)
	{
	  in >> a; in >> b; z=complex(a,b); v.Matrix[j][k]=z;
	  in >> ch;
	  if(ch!='i')
	    std::cerr << "WARNING!  Error in complex format." << std::endl;
	}
  return v;
};





#endif
