/* Input-output routines for two-qubit unitary circuit synth.               */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Standard and  Technology,                         */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */
#include<list>
#include "gatecount.h"

#ifndef _DDF_
#include<iostream>
#include<sstream>
#include<string>
using std::string;

/* Needed to detect when NaN is returned by floating point.*/
string convertToString(double x)
{
   std::ostringstream o;
   if (o << x)
     return o.str();
   // some sort of error handling goes here...
   return "conversion error";
};
#endif


string convert_int_to_string ( int n)
{ 
  std::ostringstream o;
  if( o << n)
    return o.str();
  return "string conversion errror";
};

const int DIGITS=3; /*# of digits for output real, imaginary parts of z's.*/

/* Print complexes to DIGITS.  */
string Print_complex (complex z)
{
  string s;
  if(z.im<0)
    {
      s += convertToString(Round(DIGITS, (z.re))); s += "-";
      s += convertToString(Round(DIGITS, (-1)*(z.im))); s+= "i";
    }
  else
    {
      s += convertToString(Round(DIGITS, (z.re))); s+="+";
      s += convertToString(Round(DIGITS, (z.im))); s+="i";
    };
  return s;
};

/* Print complexes to DIGITS--.  */
string Rough_Print_complex (complex z)
{
  string s;
  if(z.im<0)
    {
      s += convertToString(Round(DIGITS-1, (z.re))); s += "-";
      s += convertToString(Round(DIGITS-1, (-1)*(z.im))); s+= "i";
    }
  else
    {
      s += convertToString(Round(DIGITS-1, (z.re))); s+="+";
      s += convertToString(Round(DIGITS-1, (z.im))); s+="i";
    };
  return s;
};



/* Prints opening data, circuit box def's in LaTeX file.                 */
string LaTeX_header (void)
{
  string s;
  s += "\\documentclass[10pt]{letter}"; s+='\n'; 
  s += "\\usepackage{graphics}"; s+='\n';
  s += "\\pagestyle{empty}"; s+='\n'; s+='\n'; 
  s += "\\usepackage{amsfonts,amssymb,latexsym,amsmath,amscd}"; s+='\n'; 
  s += "\\usepackage{times,mathptm}";  s+='\n';
  s += "\\setlength{\\textheight}{9in}";  s+='\n'; 
  s += "\\renewcommand{\\thefootnote}{\\fnsymbol{footnote}}";  s+='\n';  
  s+='\n';
  
  s += "\\setlength{\\unitlength}{1.0cm}";  s+='\n';  s+='\n';

  s += "\\begin{document}";  s+='\n';  s+='\n';
  s += "\\newsavebox{\\aGate}";  s+='\n';
  s += "\\savebox{\\aGate}(0,0){";  s+='\n'; 
  s += "\\begin{picture}(0,0)";  s+='\n';
  s += "\\put(0.5,0.5){\\framebox(1,1){$a$}}";  s+='\n'; 
  s += "\\put(0,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\put(1.5,1){\\line(1,0){0.5}}";  s+='\n';
  s += "\\end{picture}"; s+='\n'; 
  s += "}";  s+='\n'; 

  s += "\\newsavebox{\\bGate}"; s+='\n'; 
  s += "\\savebox{\\bGate}(0,0){";  s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(0.5,0.5){\\framebox(1,1){$b'$}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\put(1.5,1){\\line(1,0){0.5}}";  s+='\n';
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n'; 

  s += "\\newsavebox{\\cGate}"; s+='\n'; 
  s += "\\savebox{\\cGate}(0,0){";  s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(0.5,0.5){\\framebox(1,1){$c$}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\put(1.5,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n';

  s += "\\newsavebox{\\dGate}"; s+='\n'; 
  s += "\\savebox{\\dGate}(0,0){"; s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(0.5,0.5){\\framebox(1,1){$d'$}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\put(1.5,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n';  s+='\n';  s+='\n';

  s += "\\newsavebox{\\hwire}"; s+='\n';
  s += "\\savebox{\\hwire}(0,0)"; s+='\n'; 
  s += "{"; s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){2}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n';  s+='\n'; 
    
  s += "\\newsavebox{\\botCNOT} %conditioned on bottom"; s+='\n'; 
  s += "\\savebox{\\botCNOT}(0,0){"; s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(1,3){\\circle{1}}"; s+='\n'; 
  s += "\\put(0,3){\\line(1,0){2}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){2}}"; s+='\n'; 
  s += "\\put(1,1){\\line(0,1){2.5}}"; s+='\n'; 
  s += "\\put(1,1){\\circle*{.4}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n';  s+='\n'; 
    
  s += "\\newsavebox{\\topCNOT} %conditioned on top."; s+='\n'; 
  s += "\\savebox{\\topCNOT}(0,0){"; s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(1,1){\\circle{1}}"; s+='\n'; 
  s += "\\put(0,3){\\line(1,0){2}}"; s+='\n'; 
  s += "\\put(1,.5){\\line(0,1){2.5}}"; s+='\n'; 
  s += "\\put(1,3){\\circle*{.4}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){2}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n';  s+='\n'; 

  s += "\\newsavebox{\\RzOnegate}"; s+='\n'; 
  s += "\\savebox{\\RzOnegate}(0,0){"; s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(0.5,0.5){\\framebox(1,1){$R_z(\\alpha_1)$}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\put(1.5,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n';  s+='\n'; 

  s += "\\newsavebox{\\RyOnegate}"; s+='\n'; 
  s += "\\savebox{\\RyOnegate}(0,0){"; s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(0.5,0.5){\\framebox(1,1){$R_y(\\theta_1)$}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\put(1.5,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n'; 
  s += "\\newsavebox{\\RyTwogate}"; s+='\n'; 
  s += "\\savebox{\\RyTwogate}(0,0){"; s+='\n'; 
  s += "\\begin{picture}(0,0)"; s+='\n'; 
  s += "\\put(0.5,0.5){\\framebox(1,1){$R_y(\\theta_2)$}}"; s+='\n'; 
  s += "\\put(0,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\put(1.5,1){\\line(1,0){0.5}}"; s+='\n'; 
  s += "\\end{picture}"; s+='\n'; 
  s += "}"; s+='\n';  s+='\n';  s+='\n'; 
  

  s += "\\begin{center}"; s+='\n'; 
  s += "{\\includegraphics{NISTlogo.eps}}"; s+='\n'; 
  s += "{\\tt \\LARGE Two-qubit Circuit Synth\\footnote{"; s+='\n'; 
  s += "\\begin{center}"; s+='\n'; 
  s += "Mathematical and Computational Sciences Division (891) \\\\"; s+='\n'; 
  s += "Information Technology Lab \\\\"; s+='\n'; 
  s += "National Institute of Standards and Technology, \\\\"; s+='\n'; 
  s += "Gaithersburg, MD USA \\\\"; s+='\n'; 
  s += "\\end{center}"; s+='\n';  s+='\n'; 
   
  s += "This software was developed at the National Institute of Standards";
  s += " and"; s+='\n'; 
  s += "Technology (NIST) by employees of the Federal Government in the ";
  s += "course"; s+='\n';
  s += "of their official duties. Pursuant to title 17 Section 105 of the";
  s += '\n';        
  s += "United States Code, this software is not subject to copyright";
  s += " protection"; s+='\n'; 
  s += "and is in the public domain. NIST assumes no responsibility ";
  s += "whatsoever"; s+='\n'; 
  s += "for its use by other parties, and makes no guarantees, expressed or";
  s += '\n'; 
  s += "implied about its quality, reliability, or any other characteristic.";
  s+='\n';  s+='\n'; 
    
  s += "\\noindent"; s+='\n'; 
  s += "See {\\tt q-ph/0308033}.  {\\bf  Algorithm optimized for generic";
  s+='\n'; 
  s += "unitaries requiring 3 CNOT's.}  Circuit for CNOT, $I$ will"; s+='\n';
  s += "be inefficient."; s+='\n'; s+='\n'; 

  s += "\\noindent"; s+='\n'; 
  s += "This program occassionally makes numerical errors."; s += '\n'; 
  s += "Please report replicable bugs to"; s+='\n'; 
  s += "{\\tt Stephen.Bullock@nist.gov}."; s+='\n'; 
  s += "}"; s+='\n'; 
  s += "}"; s+='\n';  /* ends footnote in TeX.*//*Yes, that's a meta-comment.*/
  s += "\\end{center}"; s+='\n';  s+='\n'; 

  s += "\\medskip"; s+='\n';  s+='\n';  
  return s;
}; /* end intialize text.*/

/*  Prints unitary U with input file into LaTeX output.                  */
string Print_U (const FourByFour& U, string input_file)
{
  string s; complex z;
  
  s += "\\noindent";  s+='\n';
  s += "{\\bf Input Matrix:}  The following unitary matrix recovered from";
  s += '\n';
  s += input_file;
  s += "\\footnote{Program automatically applies Gram-Schmidt"; s+='\n';
  s += " orthonormalization";  s+='\n';
  s += "to input matrix.}:";  s+='\n';
  s += "\\[";  s+='\n';
  s += "U = \\left(";  s+='\n';
  s += "\\begin{array}{rrrr}";  s+='\n';
  for(int j=1;j<=4;++j)
    {
      for(int k=1;k<=3;++k)
	{
	  z=U.Matrix[j][k];
	  s += Rough_Print_complex(z);
	  s +=" &";
	};
      z=U.Matrix[j][4]; s+= Rough_Print_complex (z);
      s += " \\\\";  s+='\n';
    };
    s+='\n';
    s += "\\end{array}";  s+='\n';
    s += "\\right)";  s+='\n';
    s += "\\]";  s+='\n';
  return s;
};

string Print_OneQubitMap (const OneQubitMap & v)
{
  string s; complex z;

  s += "\\left(";  s+='\n';
  s += "\\begin{array}{rr}";  s+='\n';
  s += Print_complex(v.aa); 
  s += " &  ";
  s += Print_complex(v.ab);
  s += " \\\\";  s+='\n';
  s += Print_complex(v.ba);
  s += " &  ";
  s += Print_complex(v.bb);
  s += " \\\\ ";  s+='\n';
  s += "\\end{array}";  s+='\n';
  s += "\\right)";  s+='\n';

  return s;
};

/* Prints circuit and output into LaTeX file.  Boxes in figure require     */
/* string LaTeX_header (void) definitions.                                 */
string Print_Circ (XVIII_Decomp Circuit)
{
  string s, circuit_pic, parameters; double angle;

  int count = Gate_Count_XVIII (Circuit);
  
  s += "Circuit realizing computation in ";
  s += convert_int_to_string (count);
  s += " gates as follows:";  s+='\n';  s+='\n';

  circuit_pic += "\\begin{center}"; circuit_pic +='\n';
  circuit_pic += "\\begin{picture}(14,4)"; circuit_pic +='\n';
  circuit_pic += "\\put(2,0){\\usebox{\\botCNOT}}"; circuit_pic +='\n';
  circuit_pic += "\\put(6,0){\\usebox{\\topCNOT}}"; circuit_pic +='\n';
  circuit_pic += "\\put(8,2){\\usebox{\\hwire}}"; circuit_pic +='\n';
  circuit_pic += "\\put(10,0){\\usebox{\\botCNOT}}"; circuit_pic +='\n';

  parameters += "\\noindent";   parameters+='\n';
  parameters += "{\\bf One-Qubit Parameters:}  One qubit matrices are ";
  parameters += "as follows:";   parameters+='\n';
  parameters += "\\[";  parameters+='\n';
  parameters += "\\begin{array}{l}"; parameters +='\n';

  if(! Is_Trivial_1q ( Circuit.a))
  {
    circuit_pic += "\\put(12,0){\\usebox{\\aGate}}"; circuit_pic +='\n';
    parameters+="a="; parameters+='\n';
    parameters+=Print_OneQubitMap ( Circuit.a);
    parameters +='\n'; parameters+="\\quad \\quad \\quad";
    parameters +='\n';
  }
  else
    { 
      circuit_pic += "\\put(12,0){\\usebox{\\hwire}}";
      circuit_pic += '\n';
    };

  if(! Is_Trivial_1q ( Circuit.b))
  {
    circuit_pic += "\\put(12,2){\\usebox{\\bGate}}"; circuit_pic+='\n';
    parameters += "b'=";
    parameters +='\n'; 
    parameters += Print_OneQubitMap ( Circuit.b);
    parameters += '\n';
  }
  else
    { 
      circuit_pic += "\\put(12,2){\\usebox{\\hwire}}";
      circuit_pic += '\n';
    };

  parameters += "\\\\"; parameters += '\n';
  parameters += "\\\\"; parameters += '\n';
  parameters += "\\\\"; parameters += '\n';

  if(! Is_Trivial_1q ( Circuit.c))
  {
    circuit_pic += "\\put(0,2){\\usebox{\\cGate}}";
    circuit_pic += '\n';
    parameters += "c=";
    parameters += '\n';
    parameters += Print_OneQubitMap ( Circuit.c);
    parameters += "\\quad \\quad \\quad";
    parameters += '\n';
  }
  else
    { circuit_pic += "\\put(0,2){\\usebox{\\hwire}}"; circuit_pic+='\n';};


  if(! Is_Trivial_1q ( Circuit.d))
  {
    circuit_pic += "\\put(0,0){\\usebox{\\dGate}}";
    circuit_pic += '\n';
    parameters += "d'="; parameters += '\n';
    parameters += Print_OneQubitMap ( Circuit.d);
    parameters += '\n';
  }
  else
    { circuit_pic += "\\put(0,0){\\usebox{\\hwire}}"; circuit_pic += '\n';};

  parameters += "\\end{array}"; parameters += '\n';
  parameters += "\\]"; parameters += '\n'; parameters += '\n';
  parameters += "\\noindent"; parameters += '\n';
  parameters += "{\\bf Angle parameters:}"; parameters += '\n';
  parameters += "\\quad";

  if( count_angle (Circuit.top_Rz_angle))
    {
      angle=Circuit.top_Rz_angle;
      parameters +="$\\alpha_1 = ";
      parameters += convertToString( Round(DIGITS,angle));
      parameters += "$, \\quad";
      circuit_pic += "\\put(4,2){\\usebox{\\RzOnegate}}"; circuit_pic += '\n';
    }
  else { circuit_pic +="\\put(4,2){\\usebox{\\hwire}}"; circuit_pic += '\n'; };

  if( count_angle (Circuit.left_Ry_angle))
    {
      angle=Circuit.left_Ry_angle;
      parameters +="$\\theta_1 = ";
      parameters += convertToString( Round(DIGITS,angle));
      parameters += "$, \\quad";
      circuit_pic += "\\put(4,0){\\usebox{\\RyOnegate}}"; circuit_pic += '\n';
    }
  else { circuit_pic += "\\put(4,0){\\usebox{\\hwire}}"; circuit_pic += '\n';};

  if( count_angle (Circuit.right_Ry_angle))
    {
      angle=Circuit.right_Ry_angle;
      parameters += "$\\theta_2 = ";
      parameters += convertToString(Round(DIGITS,angle));
      parameters +="$";
      circuit_pic += "\\put(8,0){\\usebox{\\RyTwogate}}"; circuit_pic += '\n';
    }
  else { circuit_pic += "\\put(8,0){\\usebox{\\hwire}}"; circuit_pic +='\n'; };

  circuit_pic += "\\end{picture}"; circuit_pic += '\n';
  circuit_pic += "\\end{center}";
  circuit_pic += '\n'; circuit_pic += '\n';/*Ends circuit_pic. */

  parameters += '\n';  parameters += '\n';

  parameters += "\\noindent";  parameters += '\n';
  parameters += "{\\bf Global phase:}";  parameters += '\n';
  parameters += "\\quad Correct circuit computation by "; 
  parameters += "$\\mbox{e}^{i\\phi}$, $\\phi=";
  angle=arg(Circuit.global_phase);
  parameters += convertToString(Round(DIGITS, angle));
  parameters += "$"; parameters += '\n'; parameters += '\n';

  parameters += "\\noindent";
  parameters += '\n'; 
  parameters += "\\hrulefill";
  parameters += '\n';  parameters += '\n';  parameters += '\n';
  
  s += circuit_pic; s += parameters;
  return s;
};

/* Prints Bloch sphere rotation conventions and closes document.         */
string Print_Conventions (void)
{
  string s;

  s += "\\noindent"; s+= '\n'; 
  s += "{\\bf Bloch Sphere rotations:}"; s+= '\n';
  s += "$R_y(t) = \\left( \\begin{array}{rr}"; s+= '\n';
  s += "\\cos \\frac{t}{2} & \\sin \\frac{t}{2} \\\\"; s+= '\n';
  s += "-\\sin \\frac{t}{2} & \\cos \\frac{t}{2} \\\\"; s+= '\n';
  s += "\\end{array}"; s+= '\n';
  s += "\\right)$,"; s+= '\n';
  s += "$R_z(t) = \\left( \\begin{array}{rr}"; s+= '\n';
  s += "\\mbox{e}^{-it/2} & 0 \\\\"; s+= '\n';
  s += "0 & \\mbox{e}^{it/2} \\\\"; s+= '\n';
  s += "\\end{array}"; s+= '\n';
  s += "\\right)$"; s+= '\n'; s+= '\n';
  s += "\\end{document}"; s+= '\n';
  return s;
};

